/*
 * ChaosExplorer source file
 * Copyright (C) 2002-2003 by Roland Riegel <feedback@roland-riegel.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * $Id$
 *
 */

#include "mandelbrot_calc.h"

Mandelbrot::Calc::Calc( Handle<CalcInputData> input, Handle<CalcResult> output )
	: QObject(), QThread(), m_calcing( false ), m_input( input ), m_output( output )
{
	connect( m_input.get(), SIGNAL( maxIterationsChanged() ), m_output.get(), SLOT( deleteData() ) );
	connect( m_input.get(), SIGNAL( exponentChanged() ), m_output.get(), SLOT( deleteData() ) );
	connect( m_input.get(), SIGNAL( optModeChanged() ), m_output.get(), SLOT( deleteData() ) );
	connect( m_input.get(), SIGNAL( coordsChanged() ), m_output.get(), SLOT( deleteData() ) );
	connect( m_input.get(), SIGNAL( colorFadeChanged() ), SLOT( colorsChanged() ) );
	connect( m_input.get(), SIGNAL( picSizeChanged() ), m_output.get(), SLOT( deleteData() ) );
}

Mandelbrot::Calc::~Calc()
{
}

int Mandelbrot::Calc::iterateComplex( const compl& pos )
{
/*	compl cpl( 0, 0 );
	for( int n = 2; n < m_input->maxIterations(); n += 3 )
	{
		cpl = pow( cpl, m_input->exponent() ) + pos;
		cpl = pow( cpl, m_input->exponent() ) + pos;
		cpl = pow( cpl, m_input->exponent() ) + pos;

		double mag = pow( cpl.real(), 2 ) + pow( cpl.imag(), 2 );

		if( mag >= 4 ) // sqrt( mag ) >= 2
		{
*/		/*	try to restore the correct iteration count
		 *	z = a + b*i, z from C, a,b from R
		 *	|z| = | a + b*i | = sqrt( a^2 + b^2 )
		 *	| z^2 | = | ( a + b*i )^2 | = sqrt( a^4 + b^4 + 2*a^2*b^2 )
		 *
		 *	mag = a^2 + b^2
		 *	mag^2 = a^4 + b^4 + 2*a^2*b^2
		 */
/*
			if( mag >= 256 )
			{
				n -= 2;
			}
			else if( mag >= 16 )
			{
				n--;
			}
			break;
		}
	}

	return n;

*/

	compl cpl( 0, 0 );
	for( int n = 0; n < m_input->maxIterations(); n++ )
	{
		cpl = pow( cpl, m_input->exponent() ) + pos;

		double mag = pow( cpl.real(), 2 ) + pow( cpl.imag(), 2 );

		if( mag >= 4 ) // sqrt( mag ) >= 2
			break;
	}

	return n;
}

void Mandelbrot::Calc::run()
{
	m_calcing = true;
	emit started();

	m_output->resize( m_input->picSize() );

	calcPalette();

	switch( m_input->optMode() )
	{
		case CalcInputData::none:
		{
			calcUnOptimized();
			break;
		}
		case CalcInputData::interleave:
		{
			calcInterleaved();
			break;
		}
		case CalcInputData::squares:
		{
			const IntRect rect( 0, 0, m_input->picSize().width - 1, m_input->picSize().height - 1 );
			if( rect.width() > rect.height() )
			{
				calcSquares( IntRect( rect.tl.x, rect.tl.y, rect.tl.x + rect.width() / 2, rect.br.y ) );
				calcSquares( IntRect( rect.tl.x + rect.width() / 2, rect.tl.y, rect.br.x, rect.br.y ) );
			}
			else
			{
				calcSquares( IntRect( rect.tl.x, rect.tl.y, rect.br.x, rect.tl.y + rect.height() / 2 ) );
				calcSquares( IntRect( rect.tl.x, rect.tl.y + rect.height() / 2, rect.br.x, rect.br.y ) );
			}
			break;
		}
		case CalcInputData::boundary_tracing:
		{
			calcBoundaryTraced();
			break;
		}
	}

	if( m_calcing )
	{
		m_calcing = false;
		emit finished();
	}
	else
	{
		emit stopped();
	}
	
}

void Mandelbrot::Calc::calcUnOptimized()
{
	IntMatrix& matrix = m_output->resultMatrix();

	const double step_x = m_input->coords().width() / m_input->picSize().width;
	const double step_y = m_input->coords().height() / m_input->picSize().height;

	for( unsigned int y = 0; y < matrix.height() && m_calcing; y++ )
	{
		for( unsigned int x = 0; x < matrix.width(); x++ )
			matrix( x, y ) = iterateComplex( compl( m_input->coords().tl.x + x * step_x, m_input->coords().tl.y + y * step_y ) );

		emit rectPainted( IntRect( 0, y, matrix.width() - 1, y ) );
	}
}

void Mandelbrot::Calc::calcInterleaved()
{
	IntMatrix& matrix = m_output->resultMatrix();

	const double step_x = m_input->coords().width() / m_input->picSize().width;
	const double step_y = m_input->coords().height() / m_input->picSize().height;
	const unsigned int interleave_step = 5;

	for( unsigned int y = 0; y < matrix.height() && m_calcing; y++ )
	{
		const double coord_y = m_input->coords().tl.y + y * step_y;

		int mirror_y = ( -coord_y - m_input->coords().tl.y ) / step_y;

		if( m_input->coords().tl.y > 0 && coord_y < 0 && mirror_y >= 0 )
		{
			for( unsigned int x = 0; x < matrix.width(); x++ )
				matrix( x, y ) = matrix( x, mirror_y );
		}
		else
		{
			unsigned int saved_x = 0;
			matrix( saved_x, y ) = iterateComplex( compl( m_input->coords().tl.x, coord_y ) );

			while( saved_x < matrix.width() - 1 )
			{
				int& iter_saved_x = matrix( saved_x, y );

				unsigned int x = saved_x + interleave_step;
				if( x > matrix.width() - 1 )
					x = matrix.width() - 1;

				for( ; x > saved_x; x-- )
				{
					int& iter_x = matrix( x, y );
					iter_x = iterateComplex( compl( m_input->coords().tl.x + x * step_x, coord_y ) );

					if( iter_x == iter_saved_x )
						break;
				}

				for( unsigned int curr_x = saved_x + 1; curr_x < x; curr_x++ )
					matrix( curr_x, y ) = iter_saved_x;

				saved_x += interleave_step;
			}
		}

		emit rectPainted( IntRect( 0, y, matrix.width() - 1, y ) );
	}
}

void Mandelbrot::Calc::calcSquares( const IntRect& rect )
{

	if( ! m_calcing )
		return;

	IntMatrix& matrix = m_output->resultMatrix();

	const double step_x = m_input->coords().width() / m_input->picSize().width;
	const double step_y = m_input->coords().height() / m_input->picSize().height;

	if( rect.width() <= 2 || rect.height() <= 2 )
	{
		// rect is quite small, simply calc all pixels
		for( int y = rect.tl.y; y <= rect.br.y; y++ )
			for( int x = rect.tl.x; x <= rect.br.x; x++ )
				matrix( x, y ) = iterateComplex( compl( m_input->coords().tl.x + x * step_x, m_input->coords().tl.y + y * step_y ) );

		emit rectPainted( rect );
	}
	else
	{
		const double coord_tl_x = m_input->coords().tl.x + rect.tl.x * step_x;
		const double coord_tl_y = m_input->coords().tl.y + rect.tl.y * step_y;
		const double coord_br_x = m_input->coords().tl.x + rect.br.x * step_x;
		const double coord_br_y = m_input->coords().tl.y + rect.br.y * step_y;

		int& tl = matrix( rect.tl.x, rect.tl.y );
		if( tl == -1 )
			tl = iterateComplex( compl( coord_tl_x, coord_tl_y ) );

		int& tr = matrix( rect.br.x, rect.tl.y );
		if( tr == -1 )
			tr = iterateComplex( compl( coord_br_x, coord_tl_y ) );

		int& bl = matrix( rect.tl.x, rect.br.y );
		if( bl == -1 )
			bl = iterateComplex( compl( coord_tl_x, coord_br_y ) );

		int& br = matrix( rect.br.x, rect.br.y );
		if( br == -1 )
			br = iterateComplex( compl( coord_br_x, coord_br_y ) );

		if( tl != tr || tl != bl || tl != br ) // quick check with the four corners
		{
			goto divide;
		}
		else
		{
			// check top and bottom corner
			for( int x = rect.tl.x + 1; x < rect.br.x; x++ )
			{
				int& t = matrix( x, rect.tl.y );
				int& b = matrix( x, rect.br.y );

				const double coord_x = m_input->coords().tl.x + x * step_x;

				if( t == -1 )
					t = iterateComplex( compl( coord_x, coord_tl_y ) );
				if( b == -1 )
					b = iterateComplex( compl( coord_x, coord_br_y ) );

				if( tl != t || tl != b )
					goto divide;

			}

			// check left and right corner
			for( int y = rect.tl.y + 1; y < rect.br.y; y++ )
			{
				int& l = matrix( rect.tl.x, y );
				int& r = matrix( rect.br.x, y );

				const double coord_y = m_input->coords().tl.y + y * step_y;

				if( l == -1 )
					l = iterateComplex( compl( coord_tl_x, coord_y ) );
				if( r == -1 )
					r = iterateComplex( compl( coord_br_x, coord_y ) );

				if( tl != l || tl != r )
					goto divide;
			}

			// all corner pixels have the same value => fill the rect
			for( y = rect.tl.y + 1; y < rect.br.y; y++ )
				for( x = rect.tl.x + 1; x < rect.br.x; x++ )
					matrix( x, y ) = tl;

			emit rectPainted( rect );
		}
	}

return;

divide:
	//divide rect
	if( rect.width() > rect.height() )
	{
		calcSquares( IntRect( rect.tl.x, rect.tl.y, rect.tl.x + rect.width() / 2, rect.br.y ) );
		calcSquares( IntRect( rect.tl.x + rect.width() / 2, rect.tl.y, rect.br.x, rect.br.y ) );
	}
	else
	{
		calcSquares( IntRect( rect.tl.x, rect.tl.y, rect.br.x, rect.tl.y + rect.height() / 2 ) );
		calcSquares( IntRect( rect.tl.x, rect.tl.y + rect.height() / 2, rect.br.x, rect.br.y ) );
	}

}

void Mandelbrot::Calc::calcBoundaryTraced()
{
	IntMatrix& matrix = m_output->resultMatrix();
	const DoubleRect& coords = m_input->coords();

	const double step_x = m_input->coords().width() / m_input->picSize().width;
	const double step_y = m_input->coords().height() / m_input->picSize().height;

	const int y_xaxis = m_input->coords().tl.y / -step_y;

	for( unsigned int y = 0; y < matrix.height() && m_calcing; y++ )
	{
		const double coord_y = m_input->coords().tl.y + y * step_y;
		const int mirror_y = ( -coord_y - m_input->coords().tl.y ) / step_y;

		if( m_input->coords().tl.y > 0 && coord_y < 0 && mirror_y >= 0 )
		{
			for( unsigned int x = 0; x < matrix.width(); x++ )
				matrix( x, y ) = matrix( x, mirror_y );
		}
		else
		{
			for( unsigned int x = 0; x < matrix.width(); x++ )
			{
				int& n = matrix( x, y );

				if( n == -1 )
					n = iterateComplex( compl( m_input->coords().tl.x + x * step_x, m_input->coords().tl.y + y * step_y ) );
				else
					continue;

				const unsigned int right = 0;
				const unsigned int down = 1;
				const unsigned int left = 2;
				const unsigned int up = 3;

				unsigned int dir = right;

				int curr_x = x;
				int curr_y = y;

				bool fill = false;

				while(1)
				{
					// step ahead
					switch( dir )
					{
						case right:
							curr_x++;
							break;

						case down:
							curr_y++;
							break;

						case left:
							curr_x--;
							break;

						case up:
							curr_y--;
							break;
					}

					if( (int) y < y_xaxis && y_xaxis < curr_y )
					{
						// other side of the x-axis, turn right
						dir = ( dir + 1 ) % 4;
						continue;
					}

					if( curr_x < 0 || curr_x > matrix.width() - 1 || curr_y < 0 || curr_y > matrix.height() - 1 )
					{
						// offscreen, turn right
						dir = ( dir + 1 ) % 4;
						continue;
					}

					if( fill && dir == down )
					{
						int paint_x = curr_x;
						while( paint_x > 0 )
						{
							paint_x--;
							int& paint_n = matrix( paint_x, curr_y );
							if( paint_n == -1 )
								paint_n = n;
							else if( paint_n != n )
								paint_x = 0;
						}
					}

					int& n_check = matrix( curr_x, curr_y );
					if( n_check == -1 )
						n_check = iterateComplex( compl( m_input->coords().tl.x + curr_x * step_x, m_input->coords().tl.y + curr_y * step_y ) );

					if( n == n_check )
						dir = ( dir + 3 ) % 4; // turn left
					else
						dir = ( dir + 1 ) % 4; // turn right

					if(  curr_x == x && curr_y == y )
					{
						if( ! fill )
						{
							fill = true;
							dir = right;
						}
						else
						{
							break;
						}
					}
				}
			}
		}

		emit rectPainted( IntRect( 0, y, matrix.width() - 1, y ) );
	}

}

void Mandelbrot::Calc::stop()
{
	m_calcing = false;
}

bool Mandelbrot::Calc::calcing()
{
	return m_calcing;
}

void Mandelbrot::Calc::colorsChanged()
{
	calcPalette();
	emit rectPainted( IntRect( 0, 0, m_input->picSize().width - 1, m_input->picSize().height - 1 ) );
}

void Mandelbrot::Calc::calcPalette()
{
	vector<QColor>& palette = m_output->palette();

	palette.resize( m_input->maxIterations() + 1 );

	int red, green, blue;
	double pos;
	for( unsigned int i = 0; i < palette.size(); i++ )
	{
		pos = (double) i / ( palette.size() - 1 );
		m_input->colorFade().colorbypos( pos, red, green, blue );
		palette[i].setRgb( red, green, blue );
	}
}
