/*
 * ChaosExplorer source file
 * Copyright (C) 2002-2003 by Roland Riegel <feedback@roland-riegel.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * $Id$
 *
 */

#ifndef MANDELBROT_CALCINPUTDATA_H
#define MANDELBROT_CALCINPUTDATA_H

#include <qobject.h>
#include "vartypes.h"

namespace Mandelbrot
{
	class CalcInputData : public QObject
	{
		Q_OBJECT
	public:
		enum Optimization
		{
			none,
			interleave,
			squares,
			boundary_tracing
		};

		CalcInputData();

		void setPicSize( const IntSize& new_picsize );
		const IntSize& picSize() const;

		void setCoords( const DoubleRect& new_coords );
		const DoubleRect& coords() const;

		void setMaxIterations( unsigned int new_maxiterations );
		unsigned int maxIterations() const;

		void setExponent( double new_exponent );
		double exponent() const;

		void setColorFade( const ColorFade& new_colorfade );
		const ColorFade& colorFade() const;

		void setOptMode( Optimization new_optmode );
		Optimization optMode() const;

	signals:
		void picSizeChanged();
		void coordsChanged();
		void maxIterationsChanged();
		void exponentChanged();
		void colorFadeChanged();
		void optModeChanged();
		void moved( int x, int y );
		
	private:
		IntSize m_picSize;
		DoubleRect m_coords;
		unsigned int m_maxIterations;
		double m_exponent;
		ColorFade m_colorFade;
		Optimization m_optMode;
	};
}

inline Mandelbrot::CalcInputData::CalcInputData()
	: QObject(), m_picSize( 320, 240 ),
	  m_coords( -2.1, 1.1625, 1.0, -1.1625 ),
	  m_maxIterations( 50 ), m_exponent( 2.0 ),
	  m_optMode( none )
{
}

inline void Mandelbrot::CalcInputData::setPicSize( const IntSize& new_picsize )
{
	if( new_picsize == m_picSize ) return;

	m_picSize = new_picsize;
	emit picSizeChanged();
}

inline const IntSize& Mandelbrot::CalcInputData::picSize() const
{
	return m_picSize;
}

inline void Mandelbrot::CalcInputData::setCoords( const DoubleRect& new_coords )
{
	if( new_coords.width() == 0 || new_coords.height() == 0 ) return;

	m_coords = new_coords;
	emit coordsChanged();
}

inline const DoubleRect& Mandelbrot::CalcInputData::coords() const
{
	return m_coords;
}

inline void Mandelbrot::CalcInputData::setMaxIterations( unsigned int new_maxiterations )
{
	m_maxIterations = new_maxiterations;
	emit maxIterationsChanged();
}

inline unsigned int Mandelbrot::CalcInputData::maxIterations() const
{
	return m_maxIterations;
}

inline void Mandelbrot::CalcInputData::setExponent( double new_exponent )
{
	m_exponent = new_exponent;
	emit exponentChanged();
}

inline double Mandelbrot::CalcInputData::exponent() const
{
	return m_exponent;
}

inline void Mandelbrot::CalcInputData::setColorFade( const ColorFade& new_colorfade )
{
	m_colorFade = new_colorfade;
	emit colorFadeChanged();
}

inline const ColorFade& Mandelbrot::CalcInputData::colorFade() const
{
	return m_colorFade;
}

inline void Mandelbrot::CalcInputData::setOptMode( Mandelbrot::CalcInputData::Optimization new_optmode )
{
	m_optMode = new_optmode;
	emit optModeChanged();
}

inline Mandelbrot::CalcInputData::Optimization Mandelbrot::CalcInputData::optMode() const
{
	return m_optMode;
}

#endif
