/*
 * ChaosExplorer source file
 * Copyright (C) 2002-2003 by Roland Riegel <feedback@roland-riegel.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * $Id$
 *
 */

#if !defined(VARTYPES_H)
#define VARTYPES_H

#include <vector>
using std::vector;

template<class T>
class Size
{
public:
	Size( T new_width = T(), T new_height = T() ) : width( new_width ), height( new_height ) {}
	Size( const Size& size ) { width = size.width; height = size.height; }
	~Size() {}

	T width;
	T height;

};

typedef Size<int> IntSize;

template<class T>
class Point
{
public:
	Point( T new_x = T(), T new_y = T() ) : x( new_x ), y( new_y ) {}
	Point( const Size<T>& size ) : x( size.width ), y( size.height ) {}
	Point( const Point<T>& point ) { x = point.x; y = point.y; }
	~Point() {}
	
	T x;
	T y;

};

typedef Point<unsigned int> UIntPoint;
typedef Point<int> IntPoint;
typedef Point<double> DoublePoint;

template<class T>
class Rect
{
public:
	Rect( Point<T> new_tl = Point<T>(), Point<T> new_br = Point<T>() ) : tl( new_tl ), br( new_br ) {}
	Rect( T new_tl_x, T new_tl_y, T new_br_x, T new_br_y ) : tl( Point<T>( new_tl_x, new_tl_y ) ), br( Point<T>( new_br_x, new_br_y ) ) {}
	Rect( Point<T> new_tl, Size<T> new_size ) : tl( new_tl ), br( Point<T>( new_tl.x + new_size.width, new_tl.y + new_size.height ) ) {}
	Rect( Point<T> new_tl, T new_width, T new_height ) : tl( new_tl ), br( Point<T>( new_tl.x + new_width, new_tl.y + new_height ) ) {}
	Rect( const Rect<T>& rect ) { tl = rect.tl; br = rect.br; }
	~Rect() {}

	void move( T dx, T dy ) { tl.x += dx; br.x += dx; tl.y += dy; br.y += dy; }

	T width() const { return br.x - tl.x; }
	T height() const { return br.y - tl.y; }
	Size<T> size() const { return Size<T>( width(), height() ); }

	Point<T> tl;
	Point<T> br;

};

typedef Rect<unsigned int> UIntRect;
typedef Rect<int> IntRect;
typedef Rect<double> DoubleRect;

template<class T>
class Matrix
{
public:

	typedef std::vector<T>::size_type size_type;

	Matrix() : m_vector(), m_width(0) {}
	Matrix( size_type size, const T& val = T() ) : m_vector( size, val ) {}
	Matrix( size_type width, size_type height, const T& val = T() ) : m_vector( width * height, val ), m_width( width ) {}
	Matrix( const Size<size_type>& size, const T& val = T() ) : m_vector( size.width * size.height, val ), m_width( size.width ) {}
	Matrix( const Matrix<T>& matrix ) { m_vector = matrix.m_vector; m_width = matrix.m_width; }
	~Matrix() {}

	T& operator[]( size_type i ) { return m_vector[i]; }
	T operator[]( size_type i ) const { return m_vector[i]; }
	T& operator[]( const Point<size_type>& point ) { return m_vector[ point.y * m_width + point.x ]; }
	T operator[]( const Point<size_type>& point ) const { return m_vector[ point.y * m_width + point.x ]; }
	T& operator()( size_type x, size_type y ) { return m_vector[ y * m_width + x ]; }
	T operator()( size_type x, size_type y ) const { return m_vector[ y * m_width + x ]; }

	size_type size() const { return m_vector.size(); }
	size_type width() const { return m_width; }
	size_type height() const { return m_width > 0 ? m_vector.size() / m_width : 0; }

	void resize( size_type size, const T& val = T() ) { m_vector.resize( size, val ); m_width = size > 0 ? 1 : 0; }
	void resize( size_type width, size_type height, const T& val = T() ) { m_vector.resize( width * height, val ); m_width = width; }
	void resize( const Size<size_type>& size, const T& val = T() ) { m_vector.resize( size.width * size.height, val ); m_width = size.width; }

private:
	vector<T> m_vector;
	size_type m_width;

};

typedef Matrix<int> IntMatrix;

struct ColorStep
{
	ColorStep( int new_red = 0, int new_green = 0, int new_blue = 0, double new_pos = 0 ) : red( new_red ), green( new_green ), blue( new_blue ) , pos( new_pos ) {}
	int red;
	int green;
	int blue;
	double pos;
};

class ColorFade
{
public:

	typedef std::vector<ColorStep>::size_type size_type;

	ColorFade()
	{
		clear();
	}

	ColorFade( const ColorFade& colorfade )
	{
		m_colors = colorfade.m_colors;
	}

	~ColorFade()
	{
	}

	void add( const ColorStep& color )
	{
		if( m_colors.size() >= 2 )
		{
			if( color.pos < m_colors[0].pos )
			{
				m_colors.insert( m_colors.begin(), color );
			}
			else
			{
				for( unsigned int i = 0; i < m_colors.size() - 1; i++ )
					if( color.pos >= m_colors[i].pos && color.pos < m_colors[ i + 1 ].pos )
					{
						m_colors.insert( m_colors.begin() + i + 1, color );
						return;
					}
				m_colors.insert( m_colors.end(), color );
			}
		}
		else if( m_colors.size() == 1 )
		{
			m_colors.insert( color.pos >= m_colors[0].pos ? m_colors.end() : m_colors.begin(), color );
		}
		else
		{
			m_colors.insert( m_colors.begin(), color );
		}
	}

	void remove( size_type pos )
	{
		if( ! empty() )
			m_colors.erase( m_colors.begin() + pos );
	}

	ColorStep& operator[]( size_type pos )
	{
		return m_colors[ pos ];
	}

	ColorStep operator[]( size_type pos ) const
	{
		return m_colors[ pos ];
	}

	size_type size() const
	{
		return m_colors.size();
	}

	bool empty() const
	{
		return m_colors.size() <= 2;
	}

	void clear()
	{
		m_colors.clear();
		add( ColorStep( 255, 255, 255, 0 ) );
		add( ColorStep( 0, 0, 0, 1 ) );
	}

	void colorbypos( double pos, int& red, int& green, int& blue ) const
	{
		if( pos >= m_colors[ m_colors.size() - 1 ].pos )
		{
			red = m_colors[ m_colors.size() - 1 ].red;
			green = m_colors[ m_colors.size() - 1 ].green;
			blue = m_colors[ m_colors.size() - 1 ].blue;
			return;
		}
		if( pos <= m_colors[0].pos )
		{
			red = m_colors[0].red;
			green = m_colors[0].green;
			blue = m_colors[0].blue;
			return;
		}

		for( unsigned int i = 0; i < m_colors.size() - 1; i++ )
			if( pos >= m_colors[i].pos && pos <= m_colors[ i + 1 ].pos )
			{
				double percentage = ( pos - m_colors[i].pos ) / ( m_colors[ i + 1 ].pos - m_colors[i].pos );
				red =  m_colors[i].red + percentage * ( m_colors[ i + 1 ].red - m_colors[i].red );
				green =  m_colors[i].green + percentage * ( m_colors[ i + 1 ].green - m_colors[i].green );
				blue = m_colors[i].blue + percentage * ( m_colors[ i + 1 ].blue - m_colors[i].blue );
				return;
			}

		red = 0;
		green = 0;
		blue = 0;
		return;
	}

	friend bool operator==( const ColorFade& colorfade1, const ColorFade& colorfade2 );
	friend bool operator!=( const ColorFade& colorfade1, const ColorFade& colorfade2 );

private:
	vector<ColorStep> m_colors;

};

template<class T>
class Handle
{
public:
	Handle() : pointer(0), pointer_count(0)
	{
	}

	Handle( T* p ) : pointer(p), pointer_count( new int(1) )
	{
	}

	Handle( const Handle& h ) : pointer( h.pointer ), pointer_count( h.pointer_count )
	{
		( *pointer_count )++;
	}

	~Handle()
	{
		if( pointer_count && --( *pointer_count ) == 0 )
		{
			if( pointer )
				delete pointer;
			delete pointer_count;
		}
	}

	Handle& operator=( const Handle& h )
	{
		if( pointer == h.pointer )
			return *this;

		if( pointer_count && --( *pointer_count ) == 0 )
		{
			if( pointer )
				delete pointer;
			delete pointer_count;
		}

		pointer = h.pointer;
		pointer_count = h.pointer_count;
		( *pointer_count )++;
		return *this;
	}

	T* operator->() const
	{
		return pointer;
	}

	T* get()
	{
		return pointer;
	}

	operator bool() const
	{
		return pointer != 0;
	}

private:
	T* pointer;
	int* pointer_count;

};


template<class T>
bool operator==( const Size<T>& size1, const Size<T>& size2 ) { return size1.width == size2.width && size1.height == size2.height; }

template<class T>
bool operator!=( const Size<T>& size1, const Size<T>& size2 ) { return size1.width != size2.width || size1.height != size2.height; }

template<class T>
bool operator==( const Point<T>& point1, const Point<T>& point2 ) { return point1.x == point2.x && point1.y == point2.y; }

template<class T>
bool operator!=( const Point<T>& point1, const Point<T>& point2 ) { return point1.x != point2.x || point1.y != point2.y; }

template<class T>
bool operator==( const Rect<T>& rect1, const Rect<T>& rect2 ) { return rect1.tl == rect2.tl && rect1.br == rect2.br; }

template<class T>
bool operator!=( const Rect<T>& rect1, const Rect<T>& rect2 ) { return rect1.tl != rect2.tl || rect1.br != rect2.br; }

bool operator==( const ColorStep& colorstep1, const ColorStep& colorstep2 );

bool operator!=( const ColorStep& colorstep1, const ColorStep& colorstep2 );

bool operator==( const ColorFade& colorfade1, const ColorFade& colorfade2 );

bool operator!=( const ColorFade& colorfade1, const ColorFade& colorfade2 );

#endif // !defined(VARTYPES_H)
